//
//  Partition Logic
//  Copyright (C) 1998-2020 J. Andrew McLaughlin
//
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the Free
//  Software Foundation; either version 2 of the License, or (at your option)
//  any later version.
//
//  This program is distributed in the hope that it will be useful, but
//  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
//  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
//  for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//
//  partlogic.c
//

#include <libintl.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/api.h>
#include <sys/charset.h>
#include <sys/color.h>
#include <sys/env.h>
#include <sys/file.h>
#include <sys/graphic.h>
#include <sys/image.h>
#include <sys/keyboard.h>
#include <sys/lang.h>
#include <sys/paths.h>
#include <sys/user.h>
#include <sys/window.h>

#define _(string) gettext(string)

#define WELCOME			_("Welcome to Partition Logic")
#define COPYRIGHT		_("Copyright (C) 1998-2020 J. Andrew McLaughlin")
#define GPL				_( \
	"  This program is free software; you can redistribute it and/or modify  \n" \
	"  it under the terms of the GNU General Public License as published by\n" \
	"  the Free Software Foundation; either version 2 of the License, or\n" \
	"  (at your option) any later version.\n\n" \
	"  This program is distributed in the hope that it will be useful, but\n" \
	"  WITHOUT ANY WARRANTY; without even the implied warranty of\n" \
	"  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See\n" \
	"  the file /system/COPYING.txt for more details.")
#define GRAPHICSMODE	_("Boot in graphics mode")
#define OK				_("OK")
#define CANCEL			_("Cancel")
#define LANGUAGE		_("Language")

// GUI stuff
static int graphics = 0;
static int bootGraphicsMode = 0;
static objectKey window = NULL;
static objectKey welcomeLabel = NULL;
static objectKey copyrightLabel = NULL;
static objectKey okButton = NULL;
static objectKey cancelButton = NULL;
static objectKey langButton = NULL;
static image flagImage;


static void setDefaults(void)
{
	char language[6];
	char charsetName[CHARSET_NAME_LEN + 1];
	char keyMapName[KEYMAP_NAMELEN + 1];
	char keyMapFile[MAX_PATH_NAME_LENGTH + 1];

	if (getenv(ENV_LANG))
	{
		strncpy(language, getenv(ENV_LANG), (sizeof(language) - 1));
	}
	else
	{
		if (configGet(PATH_SYSTEM_CONFIG "/environment.conf", ENV_LANG,
			language, (sizeof(language) - 1)) < 0)
		{
			strcpy(language, LANG_ENGLISH);
		}

		setenv(ENV_LANG, language, 1);
	}

	// Based on the default language, try to set an appropriate character
	// set variable
	if (configGet(PATH_SYSTEM_CONFIG "/charset.conf", language, charsetName,
		CHARSET_NAME_LEN) >= 0)
	{
		setenv(ENV_CHARSET, charsetName, 1);
	}

	// Based on the default language, try to set an appropriate keymap
	// variable
	if (configGet(PATH_SYSTEM_CONFIG "/keymap.conf", language, keyMapName,
		KEYMAP_NAMELEN) >= 0)
	{
		sprintf(keyMapFile, PATH_SYSTEM_KEYMAPS "/%s.map", keyMapName);

		if (fileFind(keyMapFile, NULL) >= 0)
		{
			keyboardSetMap(keyMapFile);
			setenv(ENV_KEYMAP, keyMapName, 1);
		}
	}

	setlocale(LC_ALL, language);
	textdomain("partlogic");
}


static int loadFlagImage(const char *lang, image *img)
{
	int status = 0;
	char path[MAX_PATH_LENGTH + 1];

	sprintf(path, "%s/flag-%s.bmp", PATH_SYSTEM_LOCALE, lang);

	status = fileFind(path, NULL);
	if (status < 0)
		return (status);

	return (status = imageLoad(path, 30, 20, img));
}


static void refreshWindow(void)
{
	char title[80];

	// Re-get the language setting
	setlocale(LC_ALL, getenv(ENV_LANG));
	textdomain("partlogic");

	// Re-get the character set
	if (getenv(ENV_CHARSET))
		windowSetCharSet(window, getenv(ENV_CHARSET));

	// Refresh the 'copyright' label
	windowComponentSetData(copyrightLabel, COPYRIGHT, strlen(COPYRIGHT),
		1 /* redraw */);

	// Refresh the 'ok' button
	windowComponentSetData(okButton, OK, strlen(OK), 1 /* redraw */);

	// Refresh the 'cancel' button
	windowComponentSetData(cancelButton, CANCEL, strlen(CANCEL),
		1 /* redraw */);

	if (langButton)
	{
		// Refresh the 'language' button

		if (flagImage.data)
			imageFree(&flagImage);

		if (loadFlagImage(getenv(ENV_LANG), &flagImage) >= 0)
		{
			windowComponentSetData(langButton, &flagImage, sizeof(image),
				1 /* redraw */);
		}
	}

	// Refresh the window title
	sprintf(title, "%s %s", WELCOME, RELEASE);
	windowSetTitle(window, title);

	// Re-do the window layout
	windowLayout(window);
}


static void chooseLanguage(void)
{
	char pickedLanguage[6];
	char charsetName[CHARSET_NAME_LEN + 1];
	char keyMapName[KEYMAP_NAMELEN + 1];

	if (windowNewLanguageDialog(window, pickedLanguage) >= 0)
	{
		setenv(ENV_LANG, pickedLanguage, 1);

		// Based on the chosen language, try to set an appropriate character
		// set variable
		if (configGet(PATH_SYSTEM_CONFIG "/charset.conf", pickedLanguage,
			charsetName, CHARSET_NAME_LEN) >= 0)
		{
			setenv(ENV_CHARSET, charsetName, 1);
		}

		// Based on the chosen language, try to set an appropriate keymap
		// variable
		if (configGet(PATH_SYSTEM_CONFIG "/keymap.conf", pickedLanguage,
			keyMapName, KEYMAP_NAMELEN) >= 0)
		{
			setenv(ENV_KEYMAP, keyMapName, 1);
		}

		refreshWindow();
	}
}


static void eventHandler(objectKey key, windowEvent *event)
{
	// Check for the 'ok' button
	if ((key == okButton) && (event->type == WINDOW_EVENT_MOUSE_LEFTUP))
	{
		windowGuiStop();
	}

	// Check for the 'cancel' button
	else if ((key == cancelButton) && (event->type ==
		WINDOW_EVENT_MOUSE_LEFTUP))
	{
		systemShutdown(1 /* reboot */, 0 /* don't force */);
	}

	// Check for the 'language' button
	else if (langButton && (key == langButton) && (event->type ==
		WINDOW_EVENT_MOUSE_LEFTUP))
	{
		chooseLanguage();
	}
}


static void constructWindow(const char *welcome)
{
	// If we are in graphics mode, make a window rather than operating on the
	// command line

	int status = 0;
	color background = { COLOR_DEFAULT_DESKTOP_BLUE,
		COLOR_DEFAULT_DESKTOP_GREEN, COLOR_DEFAULT_DESKTOP_RED };
	static char *splashName = PATH_SYSTEM "/partlogic.jpg";
	image splashImage;
	objectKey buttonContainer = NULL;
	int haveLocale = 0;
	componentParameters params;

	// Create a new window
	window = windowNew(multitaskerGetCurrentProcessId(), welcome);
	if (!window)
		return;

	// No title bar or border for the login window
	windowSetHasTitleBar(window, 0);
	windowSetHasBorder(window, 0);

	// Background color same as the desktop
	windowGetColor(COLOR_SETTING_DESKTOP, &background);
	windowSetBackgroundColor(window, &background);

	memset(&params, 0, sizeof(componentParameters));
	params.gridWidth = 1;
	params.gridHeight = 1;
	params.padTop = 5;
	params.padLeft = 5;
	params.padRight = 5;
	params.orientationX = orient_left;
	params.orientationY = orient_middle;
	params.flags = (COMP_PARAMS_FLAG_CUSTOMFOREGROUND |
		COMP_PARAMS_FLAG_CUSTOMBACKGROUND);
	params.foreground = COLOR_WHITE;
	memcpy(&params.background, &background, sizeof(color));
	welcomeLabel = windowNewTextLabel(window, welcome, &params);

	params.gridY += 1;
	memcpy(&params.background, &background, sizeof(color));
	copyrightLabel = windowNewTextLabel(window, COPYRIGHT, &params);

	// Try to load a splash image to go at the top of the window
	params.orientationX = orient_center;
	memset(&splashImage, 0, sizeof(image));
	if (fileFind(splashName, NULL) >= 0)
	{
		status = imageLoad(splashName, 0, 0, &splashImage);
		if (status >= 0)
		{
			// Create an image component from it, and add it to the window
			params.gridY += 1;
			windowNewImage(window, &splashImage, draw_translucent, &params);
		}
	}

	params.gridY += 1;
	params.orientationX = orient_center;
	params.flags |= (COMP_PARAMS_FLAG_FIXEDWIDTH |
		COMP_PARAMS_FLAG_FIXEDHEIGHT);
	buttonContainer = windowNewContainer(window, "buttonContainer", &params);
	if (buttonContainer)
	{
		params.padBottom = 5;
		params.orientationX = orient_right;
		params.font = NULL;
		okButton = windowNewButton(buttonContainer, OK, NULL, &params);
		windowRegisterEventHandler(okButton, &eventHandler);
		windowComponentFocus(okButton);

		// Does the 'locale' directory exist?  Anything in it?
		if (fileFind(PATH_SYSTEM_LOCALE, NULL) >= 0)
			haveLocale = 1;

		params.gridX += 1;
		if (haveLocale)
			params.orientationX = orient_center;
		else
			params.orientationX = orient_left;
		cancelButton = windowNewButton(buttonContainer, CANCEL, NULL,
			&params);
		windowRegisterEventHandler(cancelButton, &eventHandler);

		if (haveLocale)
		{
			params.gridX += 1;
			params.orientationX = orient_left;
			if (getenv(ENV_LANG))
				status = loadFlagImage(getenv(ENV_LANG), &flagImage);
			else
				status = loadFlagImage(LANG_ENGLISH, &flagImage);

			if (status >= 0)
			{
				langButton = windowNewButton(buttonContainer, NULL,
					&flagImage, &params);
			}
			else
			{
				langButton = windowNewButton(buttonContainer, LANGUAGE, NULL,
					&params);
			}

			windowRegisterEventHandler(langButton, &eventHandler);
		}
	}

	windowSetVisible(window, 1);
}


static void setGraphics(int yes)
{
	// This allows the user to toggle text/graphics mode for the next reboot

	file theFile;
	static char *noGraphName = "/nograph";

	if (yes)
	{
		// Try to delete any 'no graphics' file
		if (fileFind(noGraphName, NULL) >= 0)
			fileDelete(noGraphName);
	}

	else
	{
		// Try to create a 'no graphics' file
		memset(&theFile, 0, sizeof(file));
		if (fileFind(noGraphName, &theFile) < 0)
		{
			fileOpen(noGraphName, (OPENMODE_WRITE | OPENMODE_CREATE),
				&theFile);
			fileClose(&theFile);
		}
	}
}


int main(void)
{
	int diskManagerPid = 0;
	int logoWidth = 0;
	int logoPad = 0;
	int logoPos = 0;
	char title[80];
	char tmpChar[80];

	// Default language, character set, etc
	setDefaults();

	graphics = graphicsAreEnabled();
	bootGraphicsMode = graphics;

	setlocale(LC_ALL, getenv(ENV_LANG));
	textdomain("partlogic");

	sprintf(title, "%s %s", WELCOME, RELEASE);
	logoWidth = (strlen(title) + 10);

	if (graphics)
	{
		constructWindow(title);
		windowGuiRun();
		windowDestroy(window);

		// Log in to the window system and run the window shell
		loaderExecProgram(windowLogin(USER_ADMIN, ""), 0 /* don't block */);
	}

	else
	{
	restart:
		textScreenClear();

		// Use some extended ascii to make it prettier in text mode.  This
		// is convoluted -- not much use trying to understand it.
		logoPad = ((((logoWidth - strlen(title)) / 2)) - 1);
		logoPos = ((80 - logoWidth) / 2);
		memset(tmpChar, ' ', logoPos);
		tmpChar[logoPos++] = 201;
		memset((tmpChar + logoPos), 205, (logoWidth - 2));
		logoPos += (logoWidth - 2);
		tmpChar[logoPos++] = 187;
		tmpChar[logoPos++] = '\0';
		printf("\n%s\n", tmpChar);
		logoPos = ((80 - logoWidth) / 2);
		memset(tmpChar, ' ', logoPos);
		tmpChar[logoPos++] = 186;
		memset((tmpChar + logoPos), ' ', logoPad);
		logoPos += logoPad;
		strncpy((tmpChar + logoPos), title, strlen(title));
		logoPos += strlen(title);
		memset((tmpChar + logoPos), ' ', logoPad);
		logoPos += logoPad;
		tmpChar[logoPos++] = 186;
		tmpChar[logoPos++] = '\0';
		printf("%s\n", tmpChar);
		logoPos = ((80 - logoWidth) / 2);
		memset(tmpChar, ' ', logoPos);
		tmpChar[logoPos++] = 200;
		memset((tmpChar + logoPos), 205, (logoWidth - 2));
		logoPos += (logoWidth - 2);
		tmpChar[logoPos++] = 188;
		tmpChar[logoPos++] = '\0';
		printf("%s\n\n", tmpChar);
		printf("%s\n\n", GPL);
		printf(_("Boot mode: %s [ Press 'G' to change at next boot ]\n\n"
			"Continue? (y/n): "),
			(bootGraphicsMode? _("GRAPHICS") : _("TEXT    ")));

		while (1)
		{
			textInputSetEcho(0);
			char c = getchar();
			textInputSetEcho(1);

			switch (c)
			{
				case 'y':
				case 'Y':
					printf("%s\n", _("Yes"));
					break;

				case 'n':
				case 'N':
					printf("%s\n", _("No"));
					systemShutdown(1 /* reboot */, 0 /* don't force */);

				case 'g':
				case 'G':
					if (bootGraphicsMode)
						bootGraphicsMode = 0;
					else
						bootGraphicsMode = 1;
					setGraphics(bootGraphicsMode);
					goto restart;
			}

			break;
		}
	}

	while (1)
	{
		// Loop here, so that if the user closes the disk manager and then
		// cancels the shutdown, we will start a new disk manager

		diskManagerPid = loaderLoadProgram(PATH_PROGRAMS "/fdisk",
			0 /* privilege */);

		if (!graphics)
		{
			// Log in to the system with the disk manager as the shell
			userLogin(USER_ADMIN, "", diskManagerPid);
		}

		// Run the disk manager and block on it
		loaderExecProgram(diskManagerPid, 1 /* block */);

		// If we return to here, the user has closed the disk manager
		if (graphics)
		{
			loaderLoadAndExec(PATH_PROGRAMS "/shutdown", 0 /* privilege */,
				1 /* block */);
		}
		else
		{
			systemShutdown(1 /* reboot */, 0 /* don't force */);
		}
	}
}

